/*
 * Copyright (c) 2021 MariaDB Corporation Ab
 * Copyright (c) 2023 MariaDB plc, Finnish Branch
 *
 * Use of this software is governed by the Business Source License included
 * in the LICENSE.TXT file and at www.mariadb.com/bsl11.
 *
 * Change Date: 2025-09-12
 *
 * On the date above, in accordance with the Business Source License, use
 * of this software will be governed by version 2 or later of the General
 * Public License.
 */

#include <maxtest/galera_cluster.hh>
#include <iostream>
#include <sstream>
#include <maxtest/log.hh>
#include <maxbase/format.hh>

using std::cout;
using std::endl;
using std::string;
using mxt::MariaDBServer;

namespace
{
const string type_galera = "galera";
const string my_nwconf_prefix = "galera";
const string my_name = "Galera-cluster";
}

namespace maxtest
{
GaleraCluster::GaleraCluster(mxt::SharedData* shared)
    : MariaDBCluster(shared, "gserver")
{
}

bool GaleraCluster::setup_replication()
{
    if (!stop_nodes())
    {
        logger().log_msgf("Failed to stop servers of %s, cannot setup replication.", name().c_str());
        return false;
    }

    std::stringstream ss;

    for (int i = 0; i < N; i++)
    {
        ss << (i == 0 ? "" : ",") << ip_private(i);
    }

    auto gcomm = ss.str();

    run_on_every_backend([&](int i){
        ssh_node(i, "echo [mysqld] > cluster_address.cnf", true);
        ssh_node_f(i, true, "echo wsrep_cluster_address=gcomm://%s >>  cluster_address.cnf", gcomm.c_str());
        ssh_node(i, "cp cluster_address.cnf /etc/my.cnf.d/", true);
        ssh_node(i, "cp cluster_address.cnf /etc/mysql/my.cnf.d/", true);

        ssh_node(i, "rm -rf /var/lib/mysql/*", true);
        backend(i)->unblock_port(4567);
        backend(i)->unblock_port(4568);
        backend(i)->unblock_port(4444);
        ssh_node(i, "mariadb-install-db --user=mysql", true);

        ssh_node_f(i,
                   true,
                   "sed -i 's/###NODE-ADDRESS###/%s/' /etc/my.cnf.d/* /etc/mysql/my.cnf.d/*;"
                   "sed -i \"s|###GALERA-LIB-PATH###|$(ls /usr/lib*/galera*/*.so)|g\" /etc/my.cnf.d/* /etc/mysql/my.cnf.d/*",
                   ip_private(i));
        return true;
    });

    printf("Starting new Galera cluster\n");
    fflush(stdout);

    // Start the first node that also starts a new cluster
    auto& main_vm = backend(0)->vm_node();
    main_vm.run_cmd_sudo("galera_new_cluster");

    for (int i = 1; i < N; i++)
    {
        if (start_node(i, "") != 0)
        {
            cout << "Failed to start node" << i << endl;
            cout << "---------- BEGIN LOGS ----------" << endl;
            cout << ssh_output("sudo journalctl -u mariadb | tail -n 50", i, true).output;
            cout << "----------- END LOGS -----------" << endl;
        }
    }

    bool rval = false;
    string script_source = mxb::string_printf("%s/galera_wait_until_ready.sh", m_test_dir.c_str());
    auto homedir = main_vm.access_homedir();
    if (main_vm.copy_to_node(script_source, homedir))
    {
        string cmd = mxb::string_printf("%s/galera_wait_until_ready.sh %s",
                                        homedir, backend(0)->socket_cmd());
        main_vm.run_cmd_sudo(cmd);
        rval = create_admin_user(0) && create_users();
    }
    return rval;
}

bool GaleraCluster::check_fix_replication()
{
    bool res = false;

    if (verbose())
    {
        printf("Checking Galera\n");
        fflush(stdout);
    }

    if (connect("") == 0)
    {
        Row r = get_row(nodes[0], "SHOW STATUS WHERE Variable_name='wsrep_cluster_size'");

        if (r.size() == 2)
        {
            if (r[1] == std::to_string(N))
            {
                res = true;
            }
            else
            {
                cout << "Expected cluster size: " << N << " Actual size: " << r[1] << endl;
            }
        }
        else
        {
            cout << "Unexpected result size: "
                 << (r.empty() ? "Empty result" : std::to_string(r.size())) << endl;
        }
    }
    else
    {
        cout << "Failed to connect to the cluster" << endl;
    }

    disconnect();

    return res;
}

std::string GaleraCluster::get_srv_cnf_filename(int node)
{
    return mxb::string_printf("galera_server%i.cnf", node + 1);
}

const std::string& GaleraCluster::type_string() const
{
    return type_galera;
}

const std::string& GaleraCluster::nwconf_prefix() const
{
    return my_nwconf_prefix;
}

const std::string& GaleraCluster::name() const
{
    return my_name;
}

bool GaleraCluster::reset_server(int i)
{
    auto* srv = backend(i);
    srv->stop_database();
    srv->cleanup_database();
    reset_server_settings(i);

    auto& vm = srv->vm_node();
    auto namec = vm.m_name.c_str();

    // Note: These should be done by MDBCI
    vm.run_cmd_sudo("test -d /etc/apparmor.d/ && "
                    "ln -s /etc/apparmor.d/usr.sbin.mysqld /etc/apparmor.d/disable/usr.sbin.mysqld && "
                    "sudo service apparmor restart && "
                    "chmod a+r -R /etc/my.cnf.d/*");

    bool rval = false;

    const char reset_db_cmd[] = "mariadb-install-db; sudo chown -R mysql:mysql /var/lib/mysql";
    logger().log_msgf("Running '%s' on '%s'", reset_db_cmd, namec);
    if (vm.run_cmd_sudo(reset_db_cmd) == 0)
    {
        rval = true;
    }
    else
    {
        logger().add_failure("'%s' failed on '%s'.", reset_db_cmd, namec);
    }

    // Cannot start server yet, Galera is not properly configured.
    return rval;
}

bool GaleraCluster::create_users()
{
    bool rval = false;
    if (create_base_users())
    {
        mxt::MariaDBUserDef galmon_user = {"galeramon", "%", "galeramon"};
        galmon_user.grants = {"SUPER, REPLICATION CLIENT ON *.*"};

        auto be = backend(0);
        bool sr = supports_require();
        if (be->create_user(galmon_user, ssl_mode(), sr)
            && be->create_user(service_user_def(), ssl_mode(), sr))
        {
            rval = true;
        }
    }
    return rval;
}

bool GaleraCluster::sync_cluster()
{
    return true;
}
}
